"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var urlencode = require("isomorphic-urlencode");

/**
 * @region 类型定义
 */

/**
 * @endregion 类型定义
 */

var RequestBuilder = function () {

  /**
   * @function 默认构造函数
   * @param scheme http 或者 https
   * @param host 请求的域名
   * @param encoding 编码方式,常用的为 utf8 或者 gbk
   */


  // cookie 参数


  // 当前编码

  // 当前协议名
  function RequestBuilder() {
    var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
        scheme = _ref.scheme,
        host = _ref.host,
        encoding = _ref.encoding;

    _classCallCheck(this, RequestBuilder);

    this.scheme = "http";
    this.host = "api.com";
    this.encoding = "utf8";
    this.path = "/";
    this.cookieParams = {};
    this._option = {};

    this.scheme = scheme || this.scheme;

    this.host = host || this.host;

    this.encoding = encoding || this.encoding;
  }

  /**
   * @function 请求头设置
   * @key 请求键名或者存放键值的对象
   * @value 请求值名
   */


  // 最后封装而成的 option


  // 请求路径


  // 当前主机


  _createClass(RequestBuilder, [{
    key: "header",
    value: function header() {
      var key = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "Accept";
      var value = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "application/json";

      // 判断当前是否存在头配置
      if (!this._option.headers) {
        this._option.headers = {};
      }

      if (typeof key === "string") {
        this._option.headers[key] = value;
      } else {
        this._option.headers = Object.assign({}, this._option.headers, key);
      }

      //为了方便链式调用
      return this;
    }

    //这里输入的path是不会进行编码的,因此不要输入一些动态参数

  }, {
    key: "get",
    value: function get() {
      var path = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "/";

      //封装请求类型
      this._method("get", path);

      //重置body，避免之前使用过post
      this._option.body = null;

      return this;
    }

    /**
     * @function 以 POST 形式发起请求
     * @param path
     * @param params
     * @param contentType json / x-www-form-urlencoded
     * @return {RequestBuilder}
     */

  }, {
    key: "post",
    value: function post() {
      var path = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "/";
      var params = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      var contentType = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "json";

      //判断是否已经封装过了请求方法
      if (!params) {
        throw new Error("请设置有效请求参数");
      }

      this._method("post", path, params, contentType);

      return this;
    }

    /**
     * @function 以PUT形式发起请求
     * @param path
     * @param params
     * @param contentType
     * @return {RequestBuilder}
     */

  }, {
    key: "put",
    value: function put() {
      var path = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "/";
      var params = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      var contentType = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "json";

      //判断是否已经封装过了请求方法
      if (!params) {
        throw new Error("请设置有效请求参数");
      }

      this._method("put", path, params, contentType);

      return this;
    }

    /**
     * @function 以 DELETE 形式发起请求
     * @param path
     * @param params
     * @param contentType
     * @return {RequestBuilder}
     */

  }, {
    key: "delete",
    value: function _delete() {
      var path = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "/";
      var params = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      var contentType = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "json";

      //判断是否已经封装过了请求方法
      if (!params) {
        throw new Error("请设置有效请求参数");
      }

      this._method("delete", path, params, contentType);

      return this;
    }
  }, {
    key: "del",
    value: function del() {
      return this.delete.apply(this, arguments);
    }

    /**
     * @function 请求路径封装，自动进行编码操作
     * @param segment
     * @return {RequestBuilder}
     */

  }, {
    key: "pathSegment",
    value: function pathSegment() {
      var segment = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "";

      if (!!segment) {
        //当segment有意义值时
        this.path = this.path + "/" + this._encode(segment);
      }

      //返回当前对象
      return this;
    }

    /**
     * @function 设置本次请求为CORS
     */

  }, {
    key: "cors",
    value: function cors() {
      var origin = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "*";

      this._option.mode = "cors";

      this.header("Origin", origin);

      return this;
    }

    /**
     * function
     * @param {*} keyOrMap * 表示自动带上所有的 cookie
     * @param {*} value 当第一个参数为字符串时，使用第二个参数的键值
     */

  }, {
    key: "cookie",
    value: function cookie(keyOrMap) {
      var value = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "value";

      if (!keyOrMap) {
        return this;
      }

      if (typeof keyOrMap === "string") {
        if (keyOrMap === "*") {
          this._option.credentials = "include";
        } else {
          this.cookieParams[keyOrMap] = value;
        }
      } else {
        this.cookieParams = Object.assign({}, this.cookieParams, keyOrMap);
      }

      return this;
    }

    /**
     * @function 仅允许对于GET动作进行缓存
     * @return {RequestBuilder}
     */

  }, {
    key: "cache",
    value: function cache() {
      var cacheControl = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "no-cache";
      var maxAge = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;

      if (this._option.method !== "get") {
        throw new Error("仅允许对 GET 请求进行缓存");
      }

      this._option.cache = cacheControl;

      return this;
    }
    /**
     * @function 进行最后的构建工作,一旦调用该函数即不可以再修改之前的配置
     * @return {Promise}
     */

  }, {
    key: "build",
    value: function build(queryParams) {
      var reset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;

      // 构造请求路径
      var packagedPath = this.scheme + "://" + this.host + this.path;

      // 设置 Cookie
      if (!!this.cookieParams && Object.keys(this.cookieParams).length > 0) {
        this.header("Cookie", this._paramsToQueryString(this.cookieParams));
      }

      // 构建查询字符串
      var queryString = !!this._paramsToQueryString(queryParams) ? "?" + this._paramsToQueryString(queryParams) : "";

      var request = {
        url: packagedPath + queryString,
        option: Object.assign({}, this._option)
      };

      // 每次 build 之后会重置对象
      this.path = "/";

      // 如果选择了重置则完全重置
      if (reset) {
        this.cookieParams = {};
        this._option = {};
      } else {
        // 重置请求体
        this._option.body = null;
      }

      return request;
    }

    /**
     * @function 封装请求类型
     * @param method
     * @param path
     * @param params
     * @param contentType 设置请求内容类型 json / x-www-form-urlencoded
     * @private
     */

  }, {
    key: "_method",
    value: function _method() {
      var method = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "get";
      var path = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "/";
      var params = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
      var contentType = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : "json";

      // 设置请求方式
      this._option.method = method;

      // 设置不同的请求体格式
      if (method !== "get") {
        // Fix，对于 GET 请求不再强制设置格式
        this.header("Content-Type", "application/" + contentType);

        if (contentType === "x-www-form-urlencoded") {
          // 根据不同的编码格式设置不同的body内容
          // 将构造好的查询字符串添加到body中
          this._option.body = this._paramsToQueryString(params);
        } else {
          // 如果是以JSON形式发起请求，则直接构造JSON字符串
          this._option.body = JSON.stringify(params);
        }
      }

      // 设置请求路径
      this.path = "" + path;
    }

    /**
     * @function 利用设置好的编码格式进行编码
     * @param str
     * @private
     */

  }, {
    key: "_encode",
    value: function _encode(str) {
      if (this.encoding === "utf8") {
        return encodeURIComponent(str);
      } else {
        return urlencode(str, this.encoding);
      }
    }

    /**
     * @function 将对象解析为字符串
     * @param {*} params
     */

  }, {
    key: "_paramsToQueryString",
    value: function _paramsToQueryString(params) {
      var queryString = "";

      // 构造查询字符串
      if (!!params && Object.keys(params).length > 0) {
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = Object.keys(params)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var key = _step.value;

            queryString += key + "=" + this._encode(params[key]) + "&";
          }

          //删除最后一个无效的`&`,以避免被误认为SQL Injection
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }

        queryString = queryString.substr(0, queryString.length - 1);
      }

      return queryString;
    }
  }]);

  return RequestBuilder;
}();

exports.default = RequestBuilder;